/****************************************************************************
*   Function:   Split(), Merge()
*   Version:    1.00
*
*   Author:     Ly Minh Tr
*   Date:       05/11/94
*   Copyright:  Copyright 1994, Tri M. Ly
*
*   Purpose:    Split and merge files.
*
*   Parameter:  <inFile> - the file to be split or merge.
*               <nPart> - # of partitions to split...(from 1 to 9)
*
*   Returns:    .T. if successful, .F. otherwise
*****************************************************************************/
#if defined(_WINDLL)
    #include <windows.h>
#endif
#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <pro_ext.h>

//
// Pseudo-variables to access parameter and define library version #
//
#define pCount      parm->pCount
#define theVal1     parm->p[0].val
#define inFile      ((char FAR *) _HandToPtr(theVal1.ev_handle))
#define outFile     ((char FAR *) _HandToPtr(outFileHndl))
#define parm1Hndl   (theVal1.ev_handle)
#define parm1Len    (theVal1.ev_length)
#define theVal2     parm->p[1].val
#define nParts      (theVal2.ev_long)
#define K_VERSION   "Split() 1.0"
#define BUFSIZE     512

//
// Function prototypes
//
void FAR _Split(ParamBlk FAR *);
void FAR _Merge(ParamBlk FAR *);
void FAR _SplitVer(void);

//
// Declare FoxInfo and FoxTable data structures
//
FoxInfo myFoxInfo[] = {
  { "SPLIT", (FPFI) _Split, 2, "C,N" },
  { "MERGE", (FPFI) _Merge, 1, "C" },
  { "SPLIT_VER", (FPFI) _SplitVer, 0, "" }
};

FoxTable _FoxTable = {
    (FoxTable FAR *) 0, sizeof( myFoxInfo) / sizeof( FoxInfo ), myFoxInfo
};

/*
*
*   _Split()
*
*   Splits a file into many partitions.
*
*/
void FAR _Split(ParamBlk FAR *parm)
{
    if ((pCount != 2) || (theVal1.ev_type != 'C') ||
        (nParts < 1) || (nParts > 9))
    {
        _Execute("WAIT WINDOW [Split() Error: Invalid parameter!]");
        _RetLogical(0);
    }
    else
    {
        unsigned nAttrib, fDate, fTime;
        int nLen, nFNameLen, nResult;
        long nSize, nRead, nCount;
        char cBuffer[BUFSIZE], *pLoc, *pFLoc;
        MHANDLE outFileHndl;
        FILE *nFileIn, *nFileOut;

        //
        // Increase string by 1 character for null character
        //
        if (!_SetHandSize(parm1Hndl, parm1Len+1))
        {
            _Error(182);                    // insufficient memory
        }

        //
        // lock memory, null terminate the string
        //
        _HLock(parm1Hndl);
        inFile[parm1Len] = '\0';

        if (inFile[parm1Len-1] == '\\')
        {
            _HUnLock(parm1Hndl);
            _RetLogical(0);
            return;
        }

        strupr(inFile);

        //
        // Fetch file name portion from inFile and get it's length.
        //
        pFLoc = inFile + parm1Len - 1;
        while ((pFLoc != inFile) && (*pFLoc != '\\') && (*pFLoc != ':'))
        {
            pFLoc--;
        }

        if (pFLoc != inFile) pFLoc++;
        nFNameLen = _StrLen(pFLoc);

        //
        // Fetch file's attribute to restore after merging
        //
        _dos_getfileattr(inFile, &nAttrib);

        //
        // Determine how much more memory to get inorder to have enough
        // for file suffix.
        //
        pLoc = strrchr(inFile, '.');
        if (pLoc == NULL)
        {
            nLen = parm1Len + 5;
        }
        else
        {
            nLen = parm1Len + 3 - (parm1Len - (pLoc - inFile + 1));
        }

        if ((outFileHndl = _AllocHand(nLen)) == 0)
        {
            _Error(182);                    // insufficient memory
        }
        _HLock(outFileHndl);

        strcpy(outFile, inFile);
        outFile[nLen-4] = '.';
        outFile[nLen-3] = '0';
        outFile[nLen-2] = '0';
        outFile[nLen-1] = '1';
        outFile[nLen  ] = '\0';

        nFileIn  = fopen(inFile, "rbc");
        nFileOut = fopen(outFile, "wbc");

        if ((nFileIn == NULL) || (nFileOut == NULL))
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            _RetLogical(0);
            return;
        }

        //
        // Fetch file's date and time stamp to restore after merging
        //
        _dos_getftime(_fileno(nFileIn), &fDate, &fTime);

        //
        // Determin each partition's size
        //
        fseek(nFileIn, 0, SEEK_END);
        nSize = ftell(nFileIn) / nParts;
        if (nSize < 1) nSize = 1;

        //
        // Write header information on split files...
        //
        // Header consists of length of file name, file name (no path),
        // number of partitions, file attribute, file date and file time.
        //
        if ((fputc(nFNameLen, nFileOut) == EOF) ||
           (fwrite(pFLoc, sizeof(char), nFNameLen, nFileOut) != nFNameLen) ||
           (fputc((int) nParts, nFileOut) == EOF) ||
           (fwrite((void *) &nAttrib, sizeof(unsigned), 1, nFileOut) != 1) ||
           (fwrite((void *) &fDate, sizeof(unsigned), 1, nFileOut) != 1) ||
           (fwrite((void *) &fTime, sizeof(unsigned), 1, nFileOut) != 1))
        {
            fclose(nFileOut);
            fclose(nFileIn);
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            _RetLogical(0);
            return;
        }

        nResult = 1;
        nRead = 0;
        fseek(nFileIn, 0, SEEK_SET);

        //
        // For each partition, read in the original file and copy to each
        // partition until all of input file is read in.
        //
        while (nParts > 0)
        {
            if (!feof(nFileIn) && (nRead <= 0))
            {
                pLoc = cBuffer;
                nRead = (long) fread(cBuffer, sizeof(char), BUFSIZE, nFileIn);
            }

            //
            // Will next line will exceed the partion limit?  If not, then
            // write it to the partition, otherwise, write only enought to
            // fill the partition and continue with next partition.
            //
            if ((nRead > 0) && ((long)(ftell(nFileOut) + nRead) <= nSize) ||
                (nParts == 1))
            {
                if (nRead > 0)
                {
                    fwrite(pLoc, sizeof(char), (int) nRead, nFileOut);
                    nRead = 0;
                }

                if (feof(nFileIn) && (nParts == 1)) nParts--;
            }
            else
            {
                //
                // For the first file, remember to substract header data
                // from file length.
                //
                if (outFile[nLen-1] == '1')
                {
                    nCount = nSize - (ftell(nFileOut) - (parm1Len + 2));
                }
                else
                {
                    nCount = nSize - ftell(nFileOut);
                }

                if ((nRead > 0) && (nCount > 0))
                {
                    fwrite(pLoc, sizeof(char),
                        ((nCount < nRead) ? (int) nCount : (int) nRead),
                        nFileOut);
                }

                pLoc += ((nCount < nRead) ? nCount : nRead);
                nRead -= ((nCount < nRead) ? nCount : nRead);

                //
                // If we're not working on the last partition, then create
                // next partition, otherwise, put remaining data into last
                // partition.
                //
                if (nParts > 1)
                {
                    fclose(nFileOut);

                    if (--nParts > 0)
                    {
                        outFile[nLen-1]++;
                        nFileOut = fopen(outFile, "wbc");

                        //
                        // Write partition header info...
                        //
                        // Partition header includes length of filename,
                        // file name and partion number.
                        //
                        if ((nFileOut == NULL) ||
                           (fputc(nFNameLen, nFileOut) == EOF) ||
                           (fwrite(pFLoc, sizeof(char), nFNameLen,
                                                    nFileOut) != nFNameLen) ||
                           (fputc((outFile[nLen-1]-'0'), nFileOut) == EOF))
                        {
                            nResult = 0;
                            break;
                        }
                    }
                }
            }
        }

        fclose(nFileOut);
        fclose(nFileIn);

        _HUnLock(parm1Hndl);
        _HUnLock(outFileHndl);
        _FreeHand(outFileHndl);
        _RetLogical(nResult);
    }

    return;
}

/*
*
*   _Merge()
*
*   Merge split files partitioned by _split().
*
*/
void FAR _Merge(ParamBlk FAR *parm)
{
    if ((pCount != 1) || (theVal1.ev_type != 'C'))
    {
        _Execute("WAIT WINDOW [Merge() Error: Invalid parameter!]");
        _RetLogical(0);
    }
    else
    {
        unsigned nAttrib, fDate, fTime;
        int nFNameLen, nLen, nCount, nRead, i, nResult, nTmp;
        char cBuffer[BUFSIZE], *pLoc;
        MHANDLE outFileHndl;
        FILE *nFileIn, *nFileOut;

        //
        // Increase string by 1 character for null character
        //
        if (!_SetHandSize(parm1Hndl, parm1Len+1))
        {
            _Error(182);                    // insufficient memory
        }

        //
        // Lock memory and null terminate the string
        //
        _HLock(parm1Hndl);

        inFile[parm1Len] = '\0';
        strupr(inFile);

        //
        // Calculate if more memory is needed to store partition info.
        //
        if ((pLoc = strrchr(inFile, '.')) == NULL)
        {
            nLen = parm1Len + 4;
        }
        else
        {
            nLen = parm1Len + 3 - (parm1Len - (pLoc - inFile + 1));
        }

        if (nLen > (parm1Len + 1))
        {
            _HUnLock(parm1Hndl);

            //
            // Reset string to nLen characters for null character and suffix
            //
            if (!_SetHandSize(parm1Hndl, nLen))
            {
                _Error(182);                    // insufficient memory
            }

            //
            // lock memory, null terminate the string
            //
            _HLock(parm1Hndl);
        }

        inFile[nLen-4] = '.';
        inFile[nLen-3] = '0';
        inFile[nLen-2] = '0';
        inFile[nLen-1] = '1';
        inFile[nLen  ] = '\0';

        if ((outFileHndl = _AllocHand(nLen)) == 0)
        {
            _Error(182);                    // insufficient memory
        }
        _HLock(outFileHndl);

        memset(outFile, '\0', nLen);
        strcpy(outFile, inFile);

        //
        // Fetch output file name
        //
        pLoc = outFile + nLen - 1;
        while ((pLoc != outFile) && (*pLoc != '\\') && (*pLoc != ':'))
        {
            pLoc--;
        }
        if (pLoc != outFile) pLoc++;

        nFileIn = fopen(inFile, "rb");
        if (nFileIn == NULL)
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            _RetLogical(0);
            return;
        }

        fseek(nFileIn, 0, SEEK_SET);

        //
        // Read in length of file name from header
        //
        nFNameLen = fgetc(nFileIn);
        if (nFNameLen <= 0)
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            fclose(nFileIn);
            _RetLogical(0);
            return;
        }

        //
        // Read in file name from header and compare with given file
        // name.
        //
        nRead = fread(pLoc, sizeof(char), nFNameLen, nFileIn);
        inFile[nLen-4] = '\0';
        if ((nRead != nFNameLen) || (strstr(outFile, inFile) == NULL))
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            fclose(nFileIn);
            _RetLogical(0);
            return;
        }
        inFile[nLen-4] = '.';

        //
        // Read in number of partitions for this set.
        //
        if ((nCount = fgetc(nFileIn)) <= 0)
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            fclose(nFileIn);
            _RetLogical(0);
            return;
        }

        //
        // Read in file attribute, file date and file time.
        //
        if ((fread((void *) &nAttrib, sizeof(unsigned), 1, nFileIn) != 1) ||
            (fread((void *) &fDate, sizeof(unsigned), 1, nFileIn) != 1) ||
            (fread((void *) &fTime, sizeof(unsigned), 1, nFileIn) != 1))
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            fclose(nFileIn);
            _RetLogical(0);
            return;
        }

        nFileOut = fopen(outFile, "wb");
        if (nFileOut == NULL)
        {
            _HUnLock(parm1Hndl);
            _HUnLock(outFileHndl);
            _FreeHand(outFileHndl);
            fclose(nFileIn);
            _RetLogical(0);
            return;
        }

        //
        // For each partition, read in and write out to original file.
        //
        nResult = 1;
        for (i=1; i<=nCount; i++)
        {
            while (!feof(nFileIn))
            {
                nRead = fread(cBuffer, sizeof(char), BUFSIZE, nFileIn);
                if ((nRead > 0) &&
                    (fwrite(cBuffer, sizeof(char), nRead, nFileOut) != nRead))
                {
                    nResult = 0;
                    i += nCount;
                    break;
                }
            }

            //
            // If we're not on the last partition, then close current
            // partition and open up next partition.
            //
            if ((i + 1) <= nCount)
            {
                fclose(nFileIn);

                inFile[nLen-1]++;
                nFileIn = fopen(inFile, "rb");

                if ((nFileIn == NULL) ||
                    (nFNameLen != fgetc(nFileIn)) ||
                    (fread(pLoc, sizeof(char), nFNameLen, nFileIn) !=
                                                                nFNameLen) ||
                    (fgetc(nFileIn) != (inFile[nLen-1]-'0')))
                {
                    nResult = 0;
                    i += nCount;
                    break;
                }

                inFile[nLen-4] = '\0';
                if (strstr(outFile, inFile) == NULL)
                {
                    nResult = 0;
                    i += nCount;
                    break;
                }
                inFile[nLen-4] = '.';
            }
        }

        //
        // Restore file's original date and time, and attribute.
        //
        _dos_setftime(_fileno(nFileOut), fDate, fTime);
        fclose(nFileOut);
        _dos_setfileattr(outFile, nAttrib);

        if (nFileIn != NULL) fclose(nFileIn);

        _HUnLock(parm1Hndl);
        _HUnLock(outFileHndl);
        _FreeHand(outFileHndl);

        _RetLogical(nResult);
    }

    return;
}

/*
*
*   _SplitVer()
*
*   Returns version number of library.
*
*/
void FAR _SplitVer(void)
{
    char *author="Ly Minh Tr";
    _RetChar(K_VERSION);
    return;
}
